#include "parser.h"
#include <stdio.h>
#include <assert.h>
#include "printer.h"

//--------------------------------------------------
// Parser
//--------------------------------------------------


//--------------------------------------------------
// CharacterBasedParser
//--------------------------------------------------


int CharacterBasedParser::nextNonBlank()
{
 int c=getChar();

 while(c!=0 && (c=='\t' || c==' ' || c=='\n'))c=getChar();

 return c;
}

  
int CharacterBasedParser::nextNonBlankDoNotGet()
{
  int k=nextNonBlank();
  ungetChar(k);
  return k;
}



bool CharacterBasedParser::isLeftBrackets(int c)
{
  return c=='(' || c=='[' || c=='{';
}


bool CharacterBasedParser::isRightBrackets(int c)
{
  return c==')' || c==']' || c=='}';
}


int CharacterBasedParser::parseChar()
{
  return getChar();
}


int CharacterBasedParser::parseInt()
{
  int ret=0;
  int c=getChar();


  while(c!=0 && (c=='\t' || c==' ' || c=='\n'))c=getChar();
  if(c=='-')return -parseInt();
  while(c!=0)
    {
      if(c>='0' && c<='9')
	{
	  ret=ret*10+(c-'0');
	  c=getChar();
	}
      else
	break;
    }
  ungetChar(c);

  return ret;
}


double CharacterBasedParser::parseFloat()
{
  double ret=0;
  static char s[64];
  int i=0;

  int c=nextNonBlank();
  while(i<64 && ((c>='0' && c<='9') || c=='.' || c=='-') )
    {
      s[i++]=c;
      c=getChar();
    }      
  assert(i<64);
  s[i]=0;

  ungetChar(c);

  sscanf(s,"%f",&ret); //does this work? Are we reading floats or doubles?

  return ret;
}


FieldElement CharacterBasedParser::parseFieldElement()
{
  //  return FieldElement(parseInt());
  int a=parseInt();
  int c=nextNonBlank();
  int b=1;
  if(c=='/')
    b=parseInt();
  else
    ungetChar(c);
    
  FieldElement A=Field::staticZHomomorphism(a);
  FieldElement B=Field::staticZHomomorphism(b);

  A*=B.inverse();

  return A;
}


bool CharacterBasedParser::isVariable(int c)
{
  return (c>='a' && c<='z') || (c>='A' && c<='Z');
}


bool CharacterBasedParser::isDigit(int c)
{
  return (c>='0' && c<='9');
}


int CharacterBasedParser::variableIndex(int c)
{
  if(c>='a' && c<='z')return c-'a';
  if(c>='A' && c<='Z')return c-'A'+26;
  assert(0);
  return 0;
}


Monomial CharacterBasedParser::parseMonomial(int dimension)
{
  /* Parses a monomial without spaces and without sign */

  Monomial m(IntegerVector((dimension==-1)?0:dimension));
  IntegerVector &ret=m.exponent;

  int c=nextNonBlank();
  //  int c=getChar();

  if(c=='1')return m;

  while(c!=0 && (c=='\t' || c==' ' || c=='\n'))c=getChar();
  while(c!=0)
    {
      //         fprintf(Stderr,"parsed: %i %c\n",c,c);
      if(isVariable(c))
	{
	  int index=variableIndex(c);
	  int exponent=1;
	  if(dimension==-1)
	    ret.grow(index+1);
	  else
	    assert(index<dimension);
	  //	  c=getChar();
	  c=nextNonBlank();
	  if(isDigit(c)) // This makes parsing of Singular polynomials possible
	    {
	      ungetChar(c);
	      c='^';
	    }
	  if(c=='^')
	    {
	      exponent=parseInt();
	      //c=getChar();
	      c=nextNonBlank();
	    }
          //          printf("Index:%i, Exponent:%i\n",index,exponent);
	  ret[index]+=exponent;
	}
      else if(c=='*')
	{
	  //	  c=getChar();
	  c=nextNonBlank();
	}
      else break;
    }
  ungetChar(c);

  return m;
}


Term CharacterBasedParser::parseTerm(int dimension)
{
  //  fprintf(Stderr,"parseTerm - Begin\n");
  int c=nextNonBlank();
  if(c=='-')
    {
      c=nextNonBlank();
      ungetChar(c);
      FieldElement k;
  k=Field::staticZHomomorphism(1);
  //      k=k.one();
      if(isDigit(c))
	{
	  k=parseFieldElement();
	}
      Monomial m=parseMonomial(dimension);
      //  fprintf(Stderr,"parseTerm - End\n");
      return Term(-k,m);
    }
  else
    {
      ungetChar(c);
      FieldElement k;
      //  fprintf(Stderr,"A\n");
  k=Field::staticZHomomorphism(1);
  //      k=k.one();
  //  fprintf(Stderr,"B\n");
      if(isDigit(c))
	{
	  k=parseFieldElement();
	}
      Monomial m=parseMonomial(dimension);
      // fprintf(Stderr,"parseTerm - End\n");
      return Term(k,m);
    }

  FieldElement k=parseFieldElement();
  Monomial m=parseMonomial(dimension);

  // fprintf(Stderr,"parseTerm :");
  // AsciiPrinter(Stderr).printFieldElement(c);
  // AsciiPrinter(Stderr).printMonomial(m);

  //  fprintf(Stderr,"parseTerm - End\n");
  return Term(k,m);
}


Polynomial CharacterBasedParser::parsePolynomial(int dimension)
{
  //  fprintf(Stderr,"parsePolynomial - Begin\n");
  bool first=true;
  Polynomial p;
  Monomial firstMonomial;

  // fprintf(Stderr,"parsePolynomialBegin\n");

  while(true)
    {
      int c=nextNonBlank();

      ungetChar(c);

      if(isDigit(c)|| c=='+' || c=='-' || isVariable(c))
        {
          if(c=='+')nextNonBlank();
	  Term t=parseTerm();
	  if(first)
	    {
	      firstMonomial=t.m;
	      first=false;
	    }
          Polynomial q=Polynomial(t);
          // AsciiPrinter(Stderr).printPolynomial(q);
          if(q.getNumberOfVariables()>p.getNumberOfVariables())p.changeNumberOfVariables(q.getNumberOfVariables());
          if(p.getNumberOfVariables()>q.getNumberOfVariables())q.changeNumberOfVariables(p.getNumberOfVariables());
          p+=q;
        }
      else
        break;
    }
  // fprintf(Stderr,"parsePolynomialEnd\n");
  // AsciiPrinter(Stderr).printPolynomial(p);
  // fprintf(Stderr,"parsePolynomialEnd\n");

  if(!first)
    {
      firstMonomial.exponent.resize(p.getNumberOfVariables());
      p.mark(firstMonomial);
      //AsciiPrinter(Stderr).printVector(firstMonomial.exponent);
    }

  //fprintf(Stderr,"parsePolynomial - End\n");
  return p;
}


PolynomialSet CharacterBasedParser::parsePolynomialSet(int dimension)
{
  PolynomialSet ret;
  int c=nextNonBlank();

  assert(isLeftBrackets(c));
  do
    {
      //  AsciiPrinter(Stderr).printPolynomialSet(ret);
      //  fprintf(Stderr,"testB-\n");
      ret.push_back(parsePolynomial(dimension));
      //  AsciiPrinter(Stderr).printPolynomialSet(ret);
      //  fprintf(Stderr,"testB-\n");
    }
  while((c=nextNonBlank())==',');
  assert(isRightBrackets(c));

  // fprintf(Stderr,"testA-\n");
  //AsciiPrinter(Stderr).printPolynomialSet(ret);
  //fprintf(Stderr,"testB-\n");

 if(dimension==-1)
   {
     ret.changeNumberOfVariables();
   }
  return ret;
}


PolynomialSetList CharacterBasedParser::parsePolynomialSetList()
{
  PolynomialSetList ret;
  int c=nextNonBlank();

  assert(isLeftBrackets(c));
  do
    {
      ret.push_back(parsePolynomialSet());
    }
  while((c=nextNonBlank())==',');
  assert(isRightBrackets(c));

  return ret;
}


IntegerVector CharacterBasedParser::parseIntegerVector()
{
  list<int> temp;
  int c=nextNonBlank();

  assert(isLeftBrackets(c));
  do
    temp.push_back(parseInt());
  while((c=nextNonBlank())==',');
  assert(isRightBrackets(c));

  IntegerVector ret(temp.size());

  {
    int j=0;
    for(list<int>::iterator i=temp.begin();i!=temp.end();i++)
      ret[j++]=*i;
  }
  
  return ret;
}


IntegerVectorList CharacterBasedParser::parseIntegerVectorList()
{
  IntegerVectorList ret;
  int c=nextNonBlank();

  assert(isLeftBrackets(c));
  do
    ret.push_back(parseIntegerVector());
  while((c=nextNonBlank())==',');
  assert(isRightBrackets(c));

  return ret;
}


//--------------------------------------------------
// FileParser
//--------------------------------------------------

int FileParser::getChar()
{
  int c=getc(f);
  //  fprintf(Stderr,"getChar c=\'%c\',%i\n",c,c);
  if(c==EOF)return 0;

  return c;
}


void FileParser::ungetChar(int c)
{
  ungetc(c,f);
}


FileParser::FileParser(FILE *f)
{
  this->f=f;
}


//--------------------------------------------------
// StringParser
//--------------------------------------------------

int StringParser::getChar()
{
  if(hasUngotten)
    {
      hasUngotten=false;
      return ungotten;
    }
  return s[index++];
}

void StringParser::ungetChar(int c)
{
  assert(hasUngotten==false);
  hasUngotten=true;
  ungotten=c;
}

StringParser::StringParser(char *s):
  index(0),
  hasUngotten(false)
{
  this->s=s;
}
